/**
 * @file logistic_mpfi.c simple test of logistic map implementation with MPFI
 *
 * Ecole Précision et reproductibilité en calcul numérique
 * @date 28/03/2013
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#include <mpfi.h>
#include <mpfi_io.h>

#define IMAX 400

int logistic1 (mpfi_ptr result, mpfr_t x0, mpfr_t a);
int logistic2 (mpfi_ptr result, mpfr_t x0, mpfr_t a);
int ilogistic3 (mpfi_ptr result, mpfi_t x, mpfr_t a);

void usage (const char *name);
void read_precision  (mpfr_prec_t *prec, int argc, char **argv);
void read_parameters (mpfr_ptr a, mpfr_ptr x0, int argc, char **argv);

int
main (int argc, char **argv)
{
  int imax;
  mpfr_prec_t prec;
  mpfr_t a;
  mpfr_t x0;
  mpfi_t x;
  
  read_precision (&prec, argc, argv);

  mpfr_set_default_prec (prec);
  mpfr_init (a);
  mpfr_init (x0);

  read_parameters (a, x0, argc, argv);

  /*
   * Note: initialization using mpfi_init2
   */
  mpfi_init2 (x, prec);

  imax = logistic1 (x, x0, a);
  printf ("logistic1 %4d ", imax);
  mpfi_out_str (stdout, 10, 10, x);
  printf("\n");

  imax = logistic2 (x, x0, a);
  {
    /*
     * Note: output using mpfr_printf
     */
    mpfr_t lower_bound;
    mpfr_t upper_bound;
    mpfr_init (lower_bound);
    mpfr_init (upper_bound);
    mpfi_get_left (lower_bound, x);
    mpfi_get_right (upper_bound, x);
    mpfr_printf ("logistic2 %4d [%Re,%Re]\n", imax, lower_bound, upper_bound);
    mpfr_clear (lower_bound);
    mpfr_clear (upper_bound);
  }

  mpfi_set_fr (x, x0);
  imax = ilogistic3 (x, x, a);
  printf ("ilogistic3 %4d ", imax);
  mpfi_out_str (stdout, 10, 10, x);
  printf("\n");


  mpfi_clear (x);
  mpfr_clear (x0);
  mpfr_clear (a);
}

/**
 * compute an enclosure of the limit of logistic map.
 *
 * Implement formula:
 *  x_{n+1} = a x_n (1 - x_n)
 *
 * TODO: complete it
 *
 * @parameter[out] x computed enclosure
 * @parameter[in]  x0 initial value
 * @parameter[in]  a logistic map parameter
 * @return number of iterations
 */
int
logistic1 (mpfi_ptr x, mpfr_t x0, mpfr_t a)
{
  /* TODO */
}


/**
 * compute an enclosure of the limit of logistic map.
 *
 * Implement formula:
 *  x_{n+1} = -a (x_{n+1} -1/2)^2 + a/4
 *
 * @parameter[out] x computed enclosure
 * @parameter[in]  x0 initial value
 * @parameter[in]  a logistic map parameter
 * @return number of iterations
 */
int
logistic2 (mpfi_ptr x, mpfr_t x0, mpfr_t a)
{
  /* TODO */
}

/**
 * compute an enclosure of the limit of logistic map.
 *
 * Implement formula:
 *  X_{n+1} =
 *   F({mx_n}) + (X_n - mx_n) * F'(X_n)
 *    a mx_n (1 - mx_n) + a (X_n - mx_n) (1 - 2 X_n)
 *
 * @parameter[out] Xn computed enclosure
 * @parameter[in]  X0 initial interval value
 * @parameter[in]  a logistic map parameter
 * @return number of iterations
 */
int
ilogistic3 (mpfi_ptr Xn, mpfi_t X0, mpfr_t a)
{
  int i;
  mpfi_t X;
  mpfi_t tmp1;
  mpfi_t tmp2;
  mpfi_t x_domain;
  mpfr_t mxn;
  
  mpfr_prec_t prec;

  prec = mpfi_get_prec (Xn);
  mpfi_init2 (X, prec);
  mpfi_init2 (tmp1, prec);
  mpfi_init2 (tmp2, prec);
  mpfi_init2 (x_domain, prec);
  mpfr_init2 (mxn, prec);

  mpfi_set (Xn, X0);
  mpfi_interv_ui (x_domain, 0, 1);

  for (i = 0; i < IMAX; i++) {
    mpfi_set (X, Xn);
    mpfi_mid (mxn, Xn);
    mpfi_set_fr (tmp1, mxn);

    mpfi_si_sub (tmp1, 1, tmp1);
    mpfi_mul_fr (tmp1, tmp1, mxn);
    mpfi_mul_fr (tmp1, tmp1, a);

    mpfi_sub_fr (tmp2, Xn, mxn);
    mpfi_mul_ui (Xn, Xn, 2);
    mpfi_ui_sub (Xn, 1, Xn);
    mpfi_mul (Xn, Xn, tmp2);
    mpfi_mul_fr (Xn, Xn, a);

    mpfi_add (Xn, tmp1, Xn);

    if (!mpfi_is_inside (Xn, x_domain)) {
      printf ("ilogistic3: diverging\n");
      break;
    }
    if (mpfi_is_inside (X, Xn) && mpfi_is_inside (Xn, X)) {
      printf ("ilogistic3: no progress\n");
      break;
    }
  }

  mpfi_clear (X);
  mpfi_clear (tmp1);
  mpfi_clear (tmp2);
  mpfi_clear (x_domain);
  mpfr_clear (mxn);

  return i;
}

void
usage (const char * name)
{
  fprintf (stderr,
           "Usage: %s <precision> [<a> <x0>]\n"
           "  precision: number of bits for calculation\n"
           "  a: logistic parameter [default: a = 3.5]\n"
           "  x0: initial point [default: x0 = 0.6]\n",
           name);
  exit (1);
}

void
read_precision (mpfr_prec_t *prec, int argc, char ** argv)
{
  if (argc < 2) {
    usage (argv[0]);
  }

  errno = 0;
  *prec = strtoul (argv[1], NULL, 10);
  if (errno != 0 || *prec < 2) {
    fprintf (stderr, "Error: Cannot set precision from \"%s\"\n", argv[1]);
    exit (1);
  }
}

void
read_parameters (mpfr_ptr a, mpfr_ptr x0, int argc, char **argv)
{
  mpfi_t a_domain;
  mpfi_t x_domain;

  if (argc < 3) {
    mpfr_set_str (a, "3.5", 10, MPFR_RNDN);
    mpfr_set_str (x0, "0.6", 10, MPFR_RNDN);
    return ;
  }

  if (argc != 4) {
    usage (argv[0]);
  }

  /*
   * Note: initialization using mpfi_init with default precision
   */
  mpfi_init (a_domain);
  mpfi_init (x_domain);

  /*
   * Note: interval assignment with string
   */
  mpfi_set_str (a_domain, "[0, 4]", 10);
  /*
   * Note: interval assignment with unsigned integers
   */
  mpfi_interv_ui (x_domain, 0, 1);

  mpfr_set_str (a, argv[2], 0, MPFR_RNDN);
  if (!mpfi_is_inside_fr (a, a_domain)) {
    fprintf (stderr, "Error: Parameter a should be between 0 and 4.\n");
    exit (1);
  }

  mpfr_set_str (x0, argv[3], 0, MPFR_RNDN);
  if (!mpfi_is_inside_fr (x0, x_domain)) {
    fprintf (stderr, "Error: Start point x0 should be in ");
    mpfi_out_str (stderr, 10, 0, x_domain);
    fprintf (stderr, "\n");
    exit (1);
  }

  mpfi_clear (a_domain);
  mpfi_clear (x_domain);
}
