/**
 * @file newton.c
 *
 * Ecole Précision et reproductibilité en calcul numérique
 * @date 28/03/2013
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#include <mpfi.h>
#include <mpfi_io.h>

void logistic_mx   (mpfi_ptr result, mpfr_t x, mpfr_t a, mpfi_t b, mpfi_t c);
void ilogistic_mx  (mpfi_ptr result, mpfi_t x, mpfr_t a, mpfi_t b, mpfi_t c);
void idlogistic_mx (mpfi_ptr result, mpfi_t x, mpfr_t a);

void set_empty (mpfi_ptr x);
void half_div (mpfi_ptr left, mpfi_ptr right, mpfi_t num, mpfi_t denom);

void usage (const char *name);
void read_precision (mpfr_prec_t *prec, int argc, char **argv);
void read_parameters (mpfr_ptr a, int argc, char **argv);

int
main (int argc, char **argv)
{
  int i;

  mpfr_prec_t prec;
  mpfr_t a;
  mpfr_t midpoint;
  mpfr_t d1;
  mpfr_t d2;
  mpfr_t threshold;
  mpfr_t epsx;
  mpfr_t epsy;

  mpfi_t b;
  mpfi_t c;
  mpfi_t x1;
  mpfi_t x2;
  mpfi_t x;
  mpfi_t y;
  mpfi_t z;

#define STACK_SIZE 16
  mpfi_t candidate[STACK_SIZE];
  int candidate_nb = 0;
  mpfi_t limit[STACK_SIZE];
  int limit_nb = 0;

  read_precision (&prec, argc, argv);
  mpfr_set_default_prec (prec);

  mpfr_inits (a, midpoint, d1, d2, threshold, epsx, epsy, (mpfr_ptr)0);
  read_parameters (a, argc, argv);

  mpfi_init (b);
  mpfi_init (c);
  mpfi_init (x1);
  mpfi_init (x2);
  mpfi_init (x);
  mpfi_init (y);
  mpfi_init (z);
  for (i = 0; i < STACK_SIZE; i++) {
    mpfi_init (candidate[i]);
    mpfi_init (limit[i]);
  }

  /* 
     looking for fixed points with interval newton algorithm,
     here f(x) = logistic(x) - x
  */

  /* INTERVAL NEWTON ALGORITHM
     Input: f, f', x0     // x0 initial search interval
     Initialization: L = {x0}, alpha = 0.75 //any value in ]0.5, 1[ is suitable

     Loop: while L not empty
       Suppress (x, L)
       m := mid(x)
       (x1, x2) := m − f({m}) / f'(x) intersects x // x1 and x2 can be empty
       if w(x1) > alpha w(x) or w(x2) > alpha w(x)
         then (x1, x2) := bisect(x)
       if x1 not empty and f(x1) contains 0
         then if w(x1)/ |mid(x1)| <= epsilon x or w(f(x1)) <= epsilon Y
                then Insert x1 in Res
                else Insert x1 in L
       same handling of x2

     Output: Res, a list of intervals that may contain the roots.
  */

  /* b encloses (a-1) / 2a */
  /* c encloses (a-1)^2 / 4a */
  mpfi_set_fr (b, a);
  mpfi_sub_ui (b, b, 1);
  mpfi_div_ui (b, b, 2);
  mpfi_sqr (c, b);
  mpfi_div_fr (b, b, a);
  mpfi_div_fr (c, c, a);

  mpfi_interv_ui (candidate[candidate_nb++], 0, 1);
  mpfr_set_ui_2exp (epsx, 1, 3-prec, MPFR_RNDN);
  mpfr_set_ui_2exp (epsy, 1, 3-prec, MPFR_RNDN);

  do {
    mpfi_set (x, candidate[--candidate_nb]);  
    mpfi_mid (midpoint, x);
    mpfi_diam (threshold, x);
    mpfr_mul_d (threshold, threshold, 0.75, MPFR_RNDN);

    logistic_mx (y, midpoint, a, b, c);
    idlogistic_mx (z, x, a);

    half_div (x1, x2, y, z);
    mpfi_fr_sub (x1, midpoint, x1);
    mpfi_intersect (x1, x1, x);
    mpfi_fr_sub (x2, midpoint, x2);
    mpfi_intersect (x2, x2, x);

    if (!mpfi_is_empty (x1)) {
      mpfi_diam_abs (d1, x1);
    }
    else {
      mpfr_set_ui (d1, 0, MPFR_RNDN);
    }
    if (!mpfi_is_empty (x2)) {
      mpfi_diam_abs (d2, x2);
    }
    else {
      mpfr_set_ui (d2, 0, MPFR_RNDN);
    }
    if (mpfr_cmp (d1, threshold) > 0 || mpfr_cmp (d2, threshold) > 0) {
      mpfi_bisect (x1, x2, x);
    }

    if (!mpfi_is_empty (x1)) {
      ilogistic_mx (y, x1, a, b, c);
      if (mpfi_has_zero (y)) {
        mpfi_diam_rel (d1, x1);
        mpfi_diam_abs (d2, y);
        if (mpfr_cmp (d1, epsx) <=0 || mpfr_cmp (d2, epsy) <= 0) {
          if (limit_nb >= STACK_SIZE) {
            printf ("Error: too much solutions.\nPrinting the %d first ones.\n",
                    limit_nb);
            break;
          }
          mpfi_set (limit[limit_nb++], x1);
        }
        else {
          mpfi_set (candidate[candidate_nb++], x1);
        }
      }
    }

    if (!mpfi_is_empty (x2)) {
      ilogistic_mx (y, x2, a, b, c);
      if (mpfi_has_zero (y)) {
        mpfi_diam_rel (d1, x2);
        mpfi_diam_abs (d2, y);
        if (mpfr_cmp (d1, epsx) <=0 || mpfr_cmp (d2, epsy) <= 0) {
          if (limit_nb >= STACK_SIZE) {
            printf ("Error: too much solutions.\nPrinting the %d first ones.\n",
                    STACK_SIZE);
            break;
          }
          mpfi_set (limit[limit_nb++], x2);
        }
        else {
          if (candidate_nb >= STACK_SIZE) {
            printf ("Error: candidates are too many.\n");
            if (limit_nb != 0) {
              printf ("Printing the %d first solutions.\n", limit_nb);
            }
            break;
          }
          mpfi_set (candidate[candidate_nb++], x2);
        }
      }
    }
  } while (candidate_nb > 0);


  if (limit_nb == 0) {
    printf ("No solution.\n");
  }
  else {
    printf ("Following intervals may contains limit points:\n");
    for (i = 0; i < limit_nb; i++) {
      mpfi_out_str (stdout, 10, 0, limit[i]);
      printf ("\n");
    }
  }

  mpfr_clears (a, midpoint, d1, d2, threshold, epsx, epsy, (mpfr_ptr)0);
  mpfi_clear (b);
  mpfi_clear (c);
  mpfi_clear (x1);
  mpfi_clear (x2);
  mpfi_clear (x);
  mpfi_clear (y);
  mpfi_clear (z);
  for (i = 0; i < STACK_SIZE; i++) {
    mpfi_clear (candidate[i]);
    mpfi_clear (limit[i]);
  }
}

/**
 * Compute an enclosure of the logistic map for mpfr input minus the input.
 *
 * Implement formula:
 *  y = -a (x - b)^2 + c
 *  with b = (a-1) / 2a
 *       c = (a-1)^2 / 4a
 *
 * @parameter[out] y computed enclosure
 * @parameter[in]  x scalar variable
 * @parameter[in]  a logistic map parameter
 * @parameter[in]  b interval containing (a-1) / 2a
 * @parameter[in]  c interval containing (a-1)^2 / 4a
 */
void
logistic_mx (mpfi_ptr y, mpfr_t x, mpfr_t a, mpfi_t b, mpfi_t c)
{
  /* TODO */
}

/**
 * compute an enclosure of the logistic map for interval input minus the
 * input.
 *
 * Implement formula:
 *  y = -a (x - b)^2 + c
 *  with b = (a-1) / 2a
 *       c = (a-1)^2 / 4a
 *
 * @parameter[out] y computed enclosure
 * @parameter[in]  x interval variable
 * @parameter[in]  a logistic map parameter
 * @parameter[in]  b interval containing (a-1) / 2a
 * @parameter[in]  c interval containing (a-1)^2 / 4a
 */
void
ilogistic_mx (mpfi_ptr y, mpfi_t x, mpfr_t a, mpfi_t b, mpfi_t c)
{
  /* TODO */
}

/**
 * compute an enclosure of the logistic map derivative minus 1.
 *
 * Implement formula:
 *  y = -2 a x + a - 1
 *
 * @parameter[out] y computed enclosure
 * @parameter[in]  x interval variable
 * @parameter[in]  a logistic map parameter
 */
void
idlogistic_mx (mpfi_ptr y, mpfi_t x, mpfr_t a)
{
  /* TODO */
}

/**
 * assign an empty set value to an interval
 *
 * @parameter[out] x interval to set
 */
void
set_empty (mpfi_ptr x)
{
  mpfi_t a;
  mpfi_init (a);
  mpfi_set_ui (a, 1);
  mpfi_interv_ui (x, 2, 3);
  mpfi_intersect (x, x, a);
  mpfi_clear (a);
}

/**
 * Compute the interval parts L and R of a quotient X/Y
 *   L = {q <= 0 so that for some x in X and y in Y, we have x = q * y}
 *   R = {q >= 0 so that for some x in X and y in Y, we have x = q * y}
 *
 * @parameter[out] left negative part L of the quotient
 * @parameter[out] right positive part R of the quotient
 * @parameter[in] X numerator
 * @parameter[in] Y denominator
 */
void
half_div (mpfi_ptr left, mpfi_ptr right, mpfi_t X, mpfi_t Y)
{
  mpfi_t Xneg;
  mpfi_t Xpos;
  mpfi_t Yneg;
  mpfi_t Ypos;
  mpfi_t t1;
  mpfi_t t2;

  mpfi_init (Xneg);
  mpfi_init (Xpos);
  mpfi_init (Yneg);
  mpfi_init (Ypos);
  mpfi_init (t1);
  mpfi_init (t2);

  /*
    t1 = (-oo, 0]   t2 = [0, +oo)
  */
  mpfi_interv_ui (t1, 0, 1);
  mpfi_log (t1, t1);
  mpfi_neg (t2, t1);

  set_empty (left);
  set_empty (right);

  /*
    Xneg = X intersect (-oo, 0]  Xpos = X intersect [0, +oo)
    Yneg = Y intersect (-oo, 0]  Ypos = Y intersect [0, +oo)
  */    
  mpfi_intersect (Xneg, X, t1);
  mpfi_intersect (Yneg, Y, t1);
  mpfi_intersect (Xpos, X, t2);
  mpfi_intersect (Ypos, Y, t2);

  /*
    L = Xneg/Ypos union Xpos/Yneg 
    R = Xneg/Yneg union Xpos/Ypos
  */
  set_empty (t1);
  set_empty (t2);

  if (!mpfi_is_empty (Xneg)) {
    if (!mpfi_is_empty (Ypos)) {
      mpfi_div (left, Xneg, Ypos);
    }
    if (!mpfi_is_empty (Yneg)) {
      mpfi_div (right, Xneg, Yneg);
    }
  }
  if (!mpfi_is_empty (Xpos)) {
    if (!mpfi_is_empty (Yneg)) {
      mpfi_div (t1, Xpos, Yneg);
    }
    if (!mpfi_is_empty (Ypos)) {
      mpfi_div (t2, Xpos, Ypos);
    }
  }

  if (!mpfi_is_empty (t1)) {
    if (!mpfi_is_empty (left)) {
      mpfi_union (left, left, t1);
    }
    else {
      mpfi_set (left, t1);
    }
  }
  if (!mpfi_is_empty (t2)) {
    if (!mpfi_is_empty (right)) {
      mpfi_union (right, right, t2);
    }
    else {
      mpfi_set (right, t2);
    }
  }

  mpfi_clear (Xneg);
  mpfi_clear (Xpos);
  mpfi_clear (Yneg);
  mpfi_clear (Ypos);
  mpfi_clear (t1);
  mpfi_clear (t2);
}

void
usage (const char * name)
{
  fprintf (stderr,
           "Usage: %s <precision> [<a>]\n"
           "  precision: number of bits for calculation\n"
           "  a: logistic parameter [default: a = 0.8]\n",
           name);
  exit (1);
}

void
read_precision (mpfr_prec_t *prec, int argc, char ** argv)
{
  if (argc < 2) {
    usage (argv[0]);
  }

  errno = 0;
  *prec = strtoul (argv[1], NULL, 10);
  if (errno != 0 || *prec < 2) {
    fprintf (stderr, "Error: Cannot set precision from \"%s\"\n", argv[1]);
    exit (1);
  }
  if (*prec < 4) {
    fprintf (stderr, "Error: Precision too low.\n", argv[1]);
    exit (1);
  }
}

void
read_parameters (mpfr_ptr a, int argc, char **argv)
{
  mpfi_t a_domain;
  mpfi_t x_domain;

  if (argc < 3) {
    mpfr_set_str (a, "0.8", 10, MPFR_RNDN);
    return ;
  }

  if (argc != 3) {
    usage (argv[0]);
  }

  mpfi_init (a_domain);
  mpfi_interv_ui (a_domain, 0, 4);

  mpfr_set_str (a, argv[2], 0, MPFR_RNDN);
  if (!mpfi_is_inside_fr (a, a_domain)) {
    fprintf (stderr, "Error: Parameter a should be between 0 and 4.\n");
    exit (1);
  }

  mpfi_clear (a_domain);
}
